@@ -1,6 +1,6 @@
 ;;; tramp-smb.el --- Tramp access functions for SMB servers -*- coding: iso-8859-1; -*-
 
-;; Copyright (C) 2002 Free Software Foundation, Inc.
+;; Copyright (C) 2002, 2003, 2004 Free Software Foundation, Inc.
 
 ;; Author: Michael Albinus <Michael.Albinus@alcatel.de>
 ;; Keywords: comm, processes
@@ -47,6 +47,11 @@
 ;; ... and add it to the method list.
 (add-to-list 'tramp-methods (cons tramp-smb-method nil))
 
+;; Add a default for `tramp-default-method-alist'. Rule: If there is
+;; a domain in USER, it must be the SMB method.
+(add-to-list 'tramp-default-method-alist
+	     (list "" "%" tramp-smb-method))
+
 ;; Add completion function for SMB method.
 (tramp-set-completion-function
  tramp-smb-method
@@ -57,7 +62,7 @@
   :group 'tramp
   :type 'string)
 
-(defconst tramp-smb-prompt "^smb: \\S-+> "
+(defconst tramp-smb-prompt "^smb: \\S-+> \\|^\\s-+Server\\s-+Comment$"
   "Regexp used as prompt in smbclient.")
 
 (defconst tramp-smb-errors
@@ -66,8 +71,8 @@
    '(; Connection error
      "Connection to \\S-+ failed"
      ; Samba
-     "ERRSRV"
      "ERRDOS"
+     "ERRSRV"
      "ERRbadfile"
      "ERRbadpw"
      "ERRfilexists"
@@ -76,13 +81,16 @@
      "ERRnosuchshare"
      ; Windows NT 4.0, Windows 5.0 (Windows 2000), Windows 5.1 (Windows XP)
      "NT_STATUS_ACCESS_DENIED"
+     "NT_STATUS_ACCOUNT_LOCKED_OUT"
      "NT_STATUS_BAD_NETWORK_NAME"
      "NT_STATUS_CANNOT_DELETE"
      "NT_STATUS_LOGON_FAILURE"
+     "NT_STATUS_NETWORK_ACCESS_DENIED"
      "NT_STATUS_NO_SUCH_FILE"
      "NT_STATUS_OBJECT_NAME_INVALID"
      "NT_STATUS_OBJECT_NAME_NOT_FOUND"
-     "NT_STATUS_SHARING_VIOLATION")
+     "NT_STATUS_SHARING_VIOLATION"
+     "NT_STATUS_WRONG_PASSWORD")
    "\\|")
   "Regexp for possible error strings of SMB servers.
 Used instead of analyzing error codes of commands.")
@@ -97,17 +105,20 @@ This variable is local to each buffer.")
 This variable is local to each buffer.")
 (make-variable-buffer-local 'tramp-smb-share-cache)
 
+(defvar tramp-smb-inodes nil
+  "Keeps virtual inodes numbers for SMB files.")
+
 ;; New handlers should be added here.
 (defconst tramp-smb-file-name-handler-alist
   '(
     ;; `access-file' performed by default handler
-    (add-name-to-file . tramp-smb-not-handled)
+    (add-name-to-file . tramp-smb-handle-copy-file) ;; we're on Windows, honey.
     ;; `byte-compiler-base-file-name' performed by default handler
     (copy-file . tramp-smb-handle-copy-file)
     (delete-directory . tramp-smb-handle-delete-directory)
     (delete-file . tramp-smb-handle-delete-file)
     ;; `diff-latest-backup-file' performed by default handler
-    ;; `directory-file-name' performed by default handler
+    (directory-file-name . tramp-handle-directory-file-name)
     (directory-files . tramp-smb-handle-directory-files)
     (directory-files-and-attributes . tramp-smb-handle-directory-files-and-attributes)
     (dired-call-process . tramp-smb-not-handled)
@@ -211,12 +222,12 @@ KEEP-DATE is not handled in case NEWNAME resides on an SMB server."
 	(error "copy-file: file %s already exists" newname))
 
 ;      (with-parsed-tramp-file-name newname nil
-      (let (user host path)
+      (let (user host localname)
 	(with-parsed-tramp-file-name newname l
-	  (setq user l-user host l-host path l-path))
+	  (setq user l-user host l-host localname l-localname))
 	(save-excursion
-	  (let ((share (tramp-smb-get-share path))
-		(file (tramp-smb-get-path path t)))
+	  (let ((share (tramp-smb-get-share localname))
+		(file (tramp-smb-get-localname localname t)))
 	    (unless share
 	      (error "Target `%s' must contain a share name" filename))
 	    (tramp-smb-maybe-open-connection user host share)
@@ -236,13 +247,13 @@ KEEP-DATE is not handled in case NEWNAME resides on an SMB server."
   (unless (file-exists-p directory)
     (error "Cannot delete non-existing directory `%s'" directory))
 ;  (with-parsed-tramp-file-name directory nil
-  (let (user host path)
+  (let (user host localname)
     (with-parsed-tramp-file-name directory l
-      (setq user l-user host l-host path l-path))
+      (setq user l-user host l-host localname l-localname))
     (save-excursion
-      (let ((share (tramp-smb-get-share path))
-	    (dir (tramp-smb-get-path (file-name-directory path) t))
-	    (file (file-name-nondirectory path)))
+      (let ((share (tramp-smb-get-share localname))
+	    (dir (tramp-smb-get-localname (file-name-directory localname) t))
+	    (file (file-name-nondirectory localname)))
 	(tramp-smb-maybe-open-connection user host share)
 	(if (and
 	     (tramp-smb-send-command user host (format "cd \"%s\"" dir))
@@ -259,13 +270,13 @@ KEEP-DATE is not handled in case NEWNAME resides on an SMB server."
   (unless (file-exists-p filename)
     (error "Cannot delete non-existing file `%s'" filename))
 ;  (with-parsed-tramp-file-name filename nil
-  (let (user host path)
+  (let (user host localname)
     (with-parsed-tramp-file-name filename l
-      (setq user l-user host l-host path l-path))
+      (setq user l-user host l-host localname l-localname))
     (save-excursion
-      (let ((share (tramp-smb-get-share path))
-	    (dir (tramp-smb-get-path (file-name-directory path) t))
-	    (file (file-name-nondirectory path)))
+      (let ((share (tramp-smb-get-share localname))
+	    (dir (tramp-smb-get-localname (file-name-directory localname) t))
+	    (file (file-name-nondirectory localname)))
 	(unless (file-exists-p filename)
 	  (error "Cannot delete non-existing file `%s'" filename))
 	(tramp-smb-maybe-open-connection user host share)
@@ -276,19 +287,19 @@ KEEP-DATE is not handled in case NEWNAME resides on an SMB server."
 	    (tramp-smb-send-command user host (format "cd \\"))
 	  ;; Error
 	  (tramp-smb-send-command user host (format "cd \\"))
-	  (error "Cannot delete file `%s'" directory))))))
+	  (error "Cannot delete file `%s'" filename))))))
 
 (defun tramp-smb-handle-directory-files
   (directory &optional full match nosort)
   "Like `directory-files' for tramp files."
   (setq directory (directory-file-name (expand-file-name directory)))
 ;  (with-parsed-tramp-file-name directory nil
-  (let (user host path)
+  (let (user host localname)
     (with-parsed-tramp-file-name directory l
-      (setq user l-user host l-host path l-path))
+      (setq user l-user host l-host localname l-localname))
     (save-excursion
-      (let* ((share (tramp-smb-get-share path))
-	     (file (tramp-smb-get-path path nil))
+      (let* ((share (tramp-smb-get-share localname))
+	     (file (tramp-smb-get-localname localname nil))
 	     (entries (tramp-smb-get-file-entries user host share file)))
 	;; Just the file names are needed
 	(setq entries (mapcar 'car entries))
@@ -298,7 +309,7 @@ KEEP-DATE is not handled in case NEWNAME resides on an SMB server."
 		(delete nil
 			(mapcar (lambda (x) (when (string-match match x) x))
 				entries))))
-	;; Make absolute paths if necessary
+	;; Make absolute localnames if necessary
 	(when full
 	  (setq entries
 		(mapcar (lambda (x)
@@ -310,53 +321,58 @@ KEEP-DATE is not handled in case NEWNAME resides on an SMB server."
 	entries))))
 
 (defun tramp-smb-handle-directory-files-and-attributes
-  (directory &optional full match nosort)
+  (directory &optional full match nosort id-format)
   "Like `directory-files-and-attributes' for tramp files."
   (mapcar
    (lambda (x)
-     (cons x (file-attributes
-	(if full x (concat (file-name-as-directory directory) x)))))
+     ;; We cannot call `file-attributes' for backward compatibility reasons.
+     ;; Its optional parameter ID-FORMAT is introduced with Emacs 21.4.
+     (cons x (tramp-smb-handle-file-attributes
+	(if full x (concat (file-name-as-directory directory) x)) id-format)))
    (directory-files directory full match nosort)))
  
-(defun tramp-smb-handle-file-attributes (filename &optional nonnumeric)
-  "Like `file-attributes' for tramp files.
-Optional argument NONNUMERIC means return user and group name
-rather than as numbers."
+(defun tramp-smb-handle-file-attributes (filename &optional id-format)
+  "Like `file-attributes' for tramp files."
 ;  (with-parsed-tramp-file-name filename nil
-  (let (user host path)
+  (let (user host localname)
     (with-parsed-tramp-file-name filename l
-      (setq user l-user host l-host path l-path))
+      (setq user l-user host l-host localname l-localname))
     (save-excursion
-      (let* ((share (tramp-smb-get-share path))
-	     (file (tramp-smb-get-path path nil))
+      (let* ((share (tramp-smb-get-share localname))
+	     (file (tramp-smb-get-localname localname nil))
 	     (entries (tramp-smb-get-file-entries user host share file))
 	     (entry (and entries
-			 (assoc (file-name-nondirectory file) entries))))
+			 (assoc (file-name-nondirectory file) entries)))
+	     (uid (if (and id-format (equal id-format 'string)) "nobody" -1))
+	     (gid (if (and id-format (equal id-format 'string)) "nogroup" -1))
+	     (inode (tramp-smb-get-inode share file))
+	     (device (tramp-get-device nil tramp-smb-method user host)))
+
 	; check result
 	(when entry
 	  (list (and (string-match "d" (nth 1 entry))
 		     t)         ;0 file type
 		-1		;1 link count
-		-1		;2 uid
-		-1		;3 gid
-		(nth 3 entry)	;4 atime
+		uid		;2 uid
+		gid		;3 gid
+		'(0 0)		;4 atime
 		(nth 3 entry)	;5 mtime
-		(nth 3 entry)	;6 ctime
+		'(0 0)		;6 ctime
 		(nth 2 entry)   ;7 size
 		(nth 1 entry)   ;8 mode
 		nil		;9 gid weird
-		-1		;10 inode number
-		-1))))))	;11 file system number
+		inode		;10 inode number
+		device))))))	;11 file system number
 
 (defun tramp-smb-handle-file-directory-p (filename)
   "Like `file-directory-p' for tramp files."
 ;  (with-parsed-tramp-file-name filename nil
-  (let 	(user host path)
+  (let 	(user host localname)
     (with-parsed-tramp-file-name filename l
-      (setq user l-user host l-host path l-path))
+      (setq user l-user host l-host localname l-localname))
     (save-excursion
-      (let* ((share (tramp-smb-get-share path))
-	     (file (tramp-smb-get-path path nil))
+      (let* ((share (tramp-smb-get-share localname))
+	     (file (tramp-smb-get-localname localname nil))
 	     (entries (tramp-smb-get-file-entries user host share file))
 	     (entry (and entries
 			 (assoc (file-name-nondirectory file) entries))))
@@ -367,12 +383,12 @@ rather than as numbers."
 (defun tramp-smb-handle-file-exists-p (filename)
   "Like `file-exists-p' for tramp files."
 ;  (with-parsed-tramp-file-name filename nil
-  (let 	(user host path)
+  (let 	(user host localname)
     (with-parsed-tramp-file-name filename l
-      (setq user l-user host l-host path l-path))
+      (setq user l-user host l-host localname l-localname))
     (save-excursion
-      (let* ((share (tramp-smb-get-share path))
-	     (file (tramp-smb-get-path path nil))
+      (let* ((share (tramp-smb-get-share localname))
+	     (file (tramp-smb-get-localname localname nil))
 	     (entries (tramp-smb-get-file-entries user host share file)))
 	(and entries
 	     (member (file-name-nondirectory file) (mapcar 'car entries))
@@ -382,8 +398,8 @@ rather than as numbers."
   "Like `file-local-copy' for tramp files."
   (with-parsed-tramp-file-name filename nil
     (save-excursion
-      (let ((share (tramp-smb-get-share path))
-	    (file (tramp-smb-get-path path t))
+      (let ((share (tramp-smb-get-share localname))
+	    (file (tramp-smb-get-localname localname t))
 	    (tmpfil (tramp-make-temp-file)))
 	(unless (file-exists-p filename)
 	  (error "Cannot make local copy of non-existing file `%s'" filename))
@@ -404,13 +420,13 @@ rather than as numbers."
 (defun tramp-smb-handle-file-name-all-completions (filename directory)
   "Like `file-name-all-completions' for tramp files."
 ;  (with-parsed-tramp-file-name directory nil
-  (let (user host path)
+  (let (user host localname)
     (with-parsed-tramp-file-name directory l
-      (setq user l-user host l-host path l-path))
+      (setq user l-user host l-host localname l-localname))
     (save-match-data
       (save-excursion
-	(let* ((share (tramp-smb-get-share path))
-	       (file (tramp-smb-get-path path nil))
+	(let* ((share (tramp-smb-get-share localname))
+	       (file (tramp-smb-get-localname localname nil))
 	       (entries (tramp-smb-get-file-entries user host share file)))
 
 	  (all-completions
@@ -433,19 +449,23 @@ rather than as numbers."
 
 (defun tramp-smb-handle-file-writable-p (filename)
   "Like `file-writable-p' for tramp files."
-;  (with-parsed-tramp-file-name filename nil
-  (let 	(user host path)
-    (with-parsed-tramp-file-name filename l
-      (setq user l-user host l-host path l-path))
-    (save-excursion
-      (let* ((share (tramp-smb-get-share path))
-	     (file (tramp-smb-get-path path nil))
-	     (entries (tramp-smb-get-file-entries user host share file))
-	     (entry (and entries
-			 (assoc (file-name-nondirectory file) entries))))
-	(and entry
-	     (string-match "w" (nth 1 entry))
-	     t)))))
+  (if (not (file-exists-p filename))
+      (let ((dir (file-name-directory filename)))
+	(and (file-exists-p dir)
+	     (file-writable-p dir)))
+;    (with-parsed-tramp-file-name filename nil
+    (let (user host localname)
+      (with-parsed-tramp-file-name filename l
+	(setq user l-user host l-host localname l-localname))
+      (save-excursion
+	(let* ((share (tramp-smb-get-share localname))
+	       (file (tramp-smb-get-localname localname nil))
+	       (entries (tramp-smb-get-file-entries user host share file))
+	       (entry (and entries
+			   (assoc (file-name-nondirectory file) entries))))
+	  (and share entry
+	       (string-match "w" (nth 1 entry))
+	       t))))))
 
 (defun tramp-smb-handle-insert-directory
   (filename switches &optional wildcard full-directory-p)
@@ -457,12 +477,12 @@ WILDCARD and FULL-DIRECTORY-P are not handled."
     ;; this function is called with a non-directory ...
     (setq filename (file-name-as-directory filename)))
 ;  (with-parsed-tramp-file-name filename nil
-  (let 	(user host path)
+  (let 	(user host localname)
     (with-parsed-tramp-file-name filename l
-      (setq user l-user host l-host path l-path))
+      (setq user l-user host l-host localname l-localname))
     (save-match-data
-      (let* ((share (tramp-smb-get-share path))
-	     (file (tramp-smb-get-path path nil))
+      (let* ((share (tramp-smb-get-share localname))
+	     (file (tramp-smb-get-localname localname nil))
 	     (entries (tramp-smb-get-file-entries user host share file)))
 
 	;; Delete dummy "" entry, useless entries
@@ -510,18 +530,18 @@ WILDCARD and FULL-DIRECTORY-P are not handled."
   (unless (file-name-absolute-p dir)
     (setq dir (concat default-directory dir)))
 ;  (with-parsed-tramp-file-name dir nil
-  (let 	(user host path)
+  (let 	(user host localname)
     (with-parsed-tramp-file-name dir l
-      (setq user l-user host l-host path l-path))
+      (setq user l-user host l-host localname l-localname))
     (save-match-data
-      (let* ((share (tramp-smb-get-share path))
+      (let* ((share (tramp-smb-get-share localname))
 	     (ldir (file-name-directory dir)))
 	;; Make missing directory parts
 	(when (and parents share (not (file-directory-p ldir)))
 	  (make-directory ldir parents))
 	;; Just do it
 	(when (file-directory-p ldir)
-	  (tramp-smb-handle-make-directory-internal dir))
+	  (make-directory-internal dir))
 	(unless (file-directory-p dir)
 	  (error "Couldn't make directory %s" dir))))))
 
@@ -529,14 +549,14 @@ WILDCARD and FULL-DIRECTORY-P are not handled."
   "Like `make-directory-internal' for tramp files."
   (setq directory (directory-file-name (expand-file-name directory)))
   (unless (file-name-absolute-p directory)
-    (setq ldir (concat default-directory directory)))
+    (setq directory (concat default-directory directory)))
 ;  (with-parsed-tramp-file-name directory nil
-  (let 	(user host path)
+  (let 	(user host localname)
     (with-parsed-tramp-file-name directory l
-      (setq user l-user host l-host path l-path))
+      (setq user l-user host l-host localname l-localname))
     (save-match-data
-      (let* ((share (tramp-smb-get-share path))
-	     (file (tramp-smb-get-path path nil)))
+      (let* ((share (tramp-smb-get-share localname))
+	     (file (tramp-smb-get-localname localname nil)))
 	(when (file-directory-p (file-name-directory directory))
 	  (tramp-smb-maybe-open-connection user host share)
 	  (tramp-smb-send-command user host (format "mkdir \"%s\"" file)))
@@ -564,12 +584,12 @@ WILDCARD and FULL-DIRECTORY-P are not handled."
 	  (error "rename-file: file %s already exists" newname))
 
 ;      (with-parsed-tramp-file-name newname nil
-      (let (user host path)
+      (let (user host localname)
 	(with-parsed-tramp-file-name newname l
-	  (setq user l-user host l-host path l-path))
+	  (setq user l-user host l-host localname l-localname))
 	(save-excursion
-	  (let ((share (tramp-smb-get-share path))
-		(file (tramp-smb-get-path path t)))
+	  (let ((share (tramp-smb-get-share localname))
+		(file (tramp-smb-get-localname localname t)))
 	    (tramp-smb-maybe-open-connection user host share)
 	    (tramp-message-for-buffer
 	     nil tramp-smb-method user host
@@ -596,12 +616,12 @@ WILDCARD and FULL-DIRECTORY-P are not handled."
                               filename))
       (error "File not overwritten")))
 ;  (with-parsed-tramp-file-name filename nil
-  (let (user host path)
+  (let (user host localname)
     (with-parsed-tramp-file-name filename l
-      (setq user l-user host l-host path l-path))
+      (setq user l-user host l-host localname l-localname))
     (save-excursion
-      (let ((share (tramp-smb-get-share path))
-	    (file (tramp-smb-get-path path t))
+      (let ((share (tramp-smb-get-share localname))
+	    (file (tramp-smb-get-localname localname t))
 	    (curbuf (current-buffer))
 	    ;; We use this to save the value of `last-coding-system-used'
 	    ;; after writing the tmp file.  At the end of the function,
@@ -649,17 +669,17 @@ WILDCARD and FULL-DIRECTORY-P are not handled."
 
 ;; Internal file name functions
 
-(defun tramp-smb-get-share (path)
-  "Returns the share name of PATH."
+(defun tramp-smb-get-share (localname)
+  "Returns the share name of LOCALNAME."
   (save-match-data
-    (when (string-match "^/?\\([^/]+\\)/" path)
-      (match-string 1 path))))
+    (when (string-match "^/?\\([^/]+\\)/" localname)
+      (match-string 1 localname))))
 
-(defun tramp-smb-get-path (path convert)
-  "Returns the file name of PATH.
+(defun tramp-smb-get-localname (localname convert)
+  "Returns the file name of LOCALNAME.
 If CONVERT is non-nil exchange \"/\" by \"\\\\\"."
   (save-match-data
-    (let ((res path))
+    (let ((res localname))
 
       (setq
        res (if (string-match "^/?[^/]+/\\(.*\\)" res)
@@ -680,16 +700,16 @@ If CONVERT is non-nil exchange \"/\" by \"\\\\\"."
 
 ;; Share names of a host are cached. It is very unlikely that the
 ;; shares do change during connection.
-(defun tramp-smb-get-file-entries (user host share path)
-  "Read entries which match PATH.
+(defun tramp-smb-get-file-entries (user host share localname)
+  "Read entries which match LOCALNAME.
 Either the shares are listed, or the `dir' command is executed.
-Only entries matching the path are returned.
-Result is a list of (PATH MODE SIZE MONTH DAY TIME YEAR)."
+Only entries matching the localname are returned.
+Result is a list of (LOCALNAME MODE SIZE MONTH DAY TIME YEAR)."
   (save-excursion
     (save-match-data
-      (let ((base (or (and (> (length path) 0)
-			   (string-match "\\([^/]+\\)$" path)
-			   (regexp-quote (match-string 1 path)))
+      (let ((base (or (and (> (length localname) 0)
+			   (string-match "\\([^/]+\\)$" localname)
+			   (regexp-quote (match-string 1 localname)))
 		      ""))
 	    res entry)
 	(set-buffer (tramp-get-buffer nil tramp-smb-method user host))
@@ -702,7 +722,7 @@ Result is a list of (PATH MODE SIZE MONTH DAY TIME YEAR)."
 	    (tramp-smb-send-command
 	     user host
 	     (format "dir %s"
-		     (if (zerop (length path)) "" (concat "\"" path "*\"")))))
+		     (if (zerop (length localname)) "" (concat "\"" localname "*\"")))))
 	  (goto-char (point-min))
 	  ;; Loop the listing
 	  (unless (re-search-forward tramp-smb-errors nil t)
@@ -714,9 +734,12 @@ Result is a list of (PATH MODE SIZE MONTH DAY TIME YEAR)."
 	    ;; Cache share entries
 	    (setq tramp-smb-share-cache res)))
 
-
 	;; Add directory itself
-	(add-to-list 'res '("" "dr-xr-xr-x" 0 (0 0)))
+	(add-to-list 'res '("" "drwxrwxrwx" 0 (0 0)))
+
+	;; There's a very strange error (debugged with XEmacs 21.4.14)
+	;; If there's no short delay, it returns nil.  No idea about
+	(when (featurep 'xemacs) (sleep-for 0.01))
 
 	;; Check for matching entries
 	(delq nil (mapcar
@@ -737,21 +760,38 @@ Result is a list of (PATH MODE SIZE MONTH DAY TIME YEAR)."
 ;; They should have the format
 ;;
 ;; \s-\{2,2}                              - leading spaces
-;; \S-\(.*\S-\)\s-*                       - file name, 32 chars, left bound
+;; \S-\(.*\S-\)\s-*                       - file name, 30 chars, left bound
+;; \s-+[ADHRSV]*                          - permissions, 7 chars, right bound
 ;; \s-                                    - space delimeter
-;; \s-*[ADHRS]*                           - permissions, 5 chars, right bound
-;; \s-                                    - space delimeter
-;; \s-*[0-9]+                             - size, 8 (Samba) or 7 (Windows)
-;;                                          chars, right bound
+;; \s-+[0-9]+                             - size, 8 chars, right bound
 ;; \s-\{2,2\}                             - space delimeter
 ;; \w\{3,3\}                              - weekday
 ;; \s-                                    - space delimeter
+;; \w\{3,3\}                              - month
+;; \s-                                    - space delimeter
 ;; [ 19][0-9]                             - day
 ;; \s-                                    - space delimeter
 ;; [0-9]\{2,2\}:[0-9]\{2,2\}:[0-9]\{2,2\} - time
 ;; \s-                                    - space delimeter
 ;; [0-9]\{4,4\}                           - year
 ;;
+;; samba/src/client.c (http://samba.org/doxygen/samba/client_8c-source.html)
+;; has function display_finfo:
+;;
+;;   d_printf("  %-30s%7.7s %8.0f  %s",
+;;            finfo->name,
+;;            attrib_string(finfo->mode),
+;;            (double)finfo->size,
+;;            asctime(LocalTime(&t)));
+;;
+;; in Samba 1.9, there's the following code:
+;;
+;;   DEBUG(0,("  %-30s%7.7s%10d  %s",
+;;  	   CNV_LANG(finfo->name),
+;;	   attrib_string(finfo->mode),
+;;	   finfo->size,
+;;	   asctime(LocalTime(&t))));
+;;
 ;; Problems:
 ;; * Modern regexp constructs, like spy groups and counted repetitions, aren't
 ;;   available in older Emacsen.
@@ -763,15 +803,15 @@ Result is a list of (PATH MODE SIZE MONTH DAY TIME YEAR)."
 (defun tramp-smb-read-file-entry (share)
   "Parse entry in SMB output buffer.
 If SHARE is result, entries are of type dir. Otherwise, shares are listed.
-Result is the list (PATH MODE SIZE MTIME)."
+Result is the list (LOCALNAME MODE SIZE MTIME)."
   (let ((line (buffer-substring (point) (tramp-point-at-eol)))
-	path mode size month day hour min sec year mtime)
+	localname mode size month day hour min sec year mtime)
 
     (if (not share)
 
 	; Read share entries
 	(when (string-match "^\\s-+\\(\\S-+\\)\\s-+Disk" line)
-	  (setq path (match-string 1 line)
+	  (setq localname (match-string 1 line)
 		mode "dr-xr-xr-x"
 		size 0))
 
@@ -811,29 +851,32 @@ Result is the list (PATH MODE SIZE MTIME)."
 
 	;; size
 	(if (string-match "\\([0-9]+\\)$" line)
-	    (setq size (match-string 1 line)
-		  line (substring line 0 (- (max 8 (1+ (length size))))))
+	    (let ((length (- (max 10 (1+ (length (match-string 1 line)))))))
+	      (setq size (string-to-number (match-string 1 line)))
+	      (when (string-match "\\([ADHRSV]+\\)" (substring line length))
+		(setq length (+ length (match-end 0))))
+	      (setq line (substring line 0 length)))
 	  (return))
 
-	;; mode
-	(if (string-match "\\(\\([ADHRS]+\\)?\\s-?\\)$" line)
+	;; mode: ARCH, DIR, HIDDEN, RONLY, SYSTEM, VOLID
+	(if (string-match "\\([ADHRSV]+\\)?$" line)
 	    (setq
-	     mode (or (match-string 2 line) "")
+	     mode (or (match-string 1 line) "")
 	     mode (save-match-data (format
 		    "%s%s"
 		    (if (string-match "D" mode) "d" "-")
 		    (mapconcat
 		     (lambda (x) "") "    "
 		     (concat "r" (if (string-match "R" mode) "-" "w") "x"))))
-	     line (substring line 0 (- (1+ (length (match-string 2 line))))))
+	     line (substring line 0 -7))
 	  (return))
 
-	;; path
-	(if (string-match "^\\s-+\\(\\S-\\(.*\\S-\\)?\\)\\s-+$" line)
-	    (setq path (match-string 1 line))
+	;; localname
+	(if (string-match "^\\s-+\\(\\S-\\(.*\\S-\\)?\\)\\s-*$" line)
+	    (setq localname (match-string 1 line))
 	  (return))))
 
-    (when (and path mode size)
+    (when (and localname mode size)
       (setq mtime
 	    (if (and sec min hour day month year)
 		(encode-time
@@ -841,7 +884,21 @@ Result is the list (PATH MODE SIZE MTIME)."
 		 (cdr (assoc (downcase month) tramp-smb-parse-time-months))
 		 year)
 	      '(0 0)))
-      (list path mode size mtime))))
+      (list localname mode size mtime))))
+
+;; Inodes don't exist for SMB files.  Therefore we must generate virtual ones.
+;; Used in `find-buffer-visiting'.
+;; The method applied might be not so efficient (Ange-FTP uses hashes). But
+;; performance isn't the major issue given that file transfer will take time.
+
+(defun tramp-smb-get-inode (share file)
+  "Returns the virtual inode number.
+If it doesn't exist, generate a new one."
+  (let ((string (concat share "/" (directory-file-name file))))
+    (unless (assoc string tramp-smb-inodes)
+      (add-to-list 'tramp-smb-inodes
+		   (list string (length tramp-smb-inodes))))
+    (nth 1 (assoc string tramp-smb-inodes))))
 
 
 ;; Connection functions
@@ -860,7 +917,8 @@ there has been an error message from smbclient."
   "Maybe open a connection to HOST, logging in as USER, using `tramp-smb-program'.
 Does not do anything if a connection is already open, but re-opens the
 connection if a previous connection has died for some reason."
-  (let ((p (get-buffer-process
+  (let ((process-connection-type tramp-process-connection-type)
+	(p (get-buffer-process
 	    (tramp-get-buffer nil tramp-smb-method user host))))
     (save-excursion
       (set-buffer (tramp-get-buffer nil tramp-smb-method user host))
@@ -943,58 +1001,44 @@ Domain names in USER and port numbers in HOST are acknowledged."
 	    (tramp-enter-password p pw-prompt)))
 
 	(unless (tramp-smb-wait-for-output user host)
+	  (tramp-clear-passwd user host)
 	  (error "Cannot open connection //%s@%s/%s"
 		 user host (or share "")))))))
 
-;; We don't use timeouts.  If needed, the caller shall warap around.
+;; We don't use timeouts.  If needed, the caller shall wrap around.
 (defun tramp-smb-wait-for-output (user host)
   "Wait for output from smbclient command.
-Sets position to begin of buffer.
-Returns nil if no error message has appeared."
-  (save-excursion
-    (let ((proc (get-buffer-process (current-buffer)))
-	  found err)
-      (save-match-data
-	;; Algorithm: get waiting output.  See if last line contains
-	;; tramp-smb-prompt sentinel, or process has exited.
-	;; If not, wait a bit and again get waiting output.
-	(while (and (not found)
-		    proc (processp proc)
-		    (memq (process-status proc) '(run open)))
-	  (accept-process-output proc 10)
-	  (goto-char (point-max))
-	  (beginning-of-line)
-	  (setq found (looking-at tramp-smb-prompt)))
+Returns nil if an error message has appeared."
+  (let ((proc (get-buffer-process (current-buffer)))
+	(found (progn (goto-char (point-min))
+		      (re-search-forward tramp-smb-prompt nil t)))
+	(err   (progn (goto-char (point-min))
+		      (re-search-forward tramp-smb-errors nil t))))
 
-	;; There might be pending output.  If tramp-smb-prompt sentinel
-	;; hasn't been found, the process has died already.  We should
-	;; give it a chance.
-	(when (not found)
-	  (if (and proc (processp proc))
-	      (accept-process-output proc 1)
-	    (accept-process-output nil 1)))
+    ;; Algorithm: get waiting output.  See if last line contains
+    ;; tramp-smb-prompt sentinel or tramp-smb-errors strings.
+    ;; If not, wait a bit and again get waiting output.
+    (while (and (not found) (not err))
 
-	;; Search for errors.
-	(goto-char (point-min))
-	(setq err (re-search-forward tramp-smb-errors nil t)))
+      ;; Accept pending output.
+      (accept-process-output proc)
 
-      ;; Add output to debug buffer if appropriate.
-      (when tramp-debug-buffer
-	(append-to-buffer
-	 (tramp-get-debug-buffer nil tramp-smb-method user host)
-	 (point-min) (point-max))
-	(when (and (not found)
-		   proc (processp proc)
-		   (memq (process-status proc) '(run open)))
-	  (save-excursion
-	    (set-buffer
-	     (tramp-get-debug-buffer nil tramp-smb-method user host))
-	    (goto-char (point-max))
-	    (insert (format "[[Remote prompt `%s' not found]]\n"
-			    tramp-smb-prompt)))))
+      ;; Search for prompt.
       (goto-char (point-min))
-      ;; Return value is whether no error message has appeared.
-      (not err))))
+      (setq found (re-search-forward tramp-smb-prompt nil t))
+
+      ;; Search for errors.
+      (goto-char (point-min))
+      (setq err (re-search-forward tramp-smb-errors nil t)))
+
+    ;; Add output to debug buffer if appropriate.
+    (when tramp-debug-buffer
+      (append-to-buffer
+       (tramp-get-debug-buffer nil tramp-smb-method user host)
+       (point-min) (point-max)))
+
+    ;; Return value is whether no error message has appeared.
+    (not err)))
 
 
 ;; Snarfed code from time-date.el and parse-time.el
@@ -1072,21 +1116,19 @@ Return the difference in the format of a time value."
 ;; * Provide a local smb.conf. The default one might not be readable.
 ;; * Error handling in case password is wrong.
 ;; * Read password from "~/.netrc".
-;; * `tramp-smb-wait-for-output' has problems with Emacs 20.7 (when process
-;;   has died).  To be investigated further.
-;; * Use different buffers for different shares.  By this, the password
-;;   won't be requested again when changing shares on the same host.
 ;; * Return more comprehensive file permission string.  Think whether it is
 ;;   possible to implement `set-file-modes'.
 ;; * Handle WILDCARD and FULL-DIRECTORY-P in
-;;   `tramp-smb-handle-insert-directory'. Remove workaround returning "".
+;;   `tramp-smb-handle-insert-directory'.
+;; * Handle links (FILENAME.LNK).
 ;; * Maybe local tmp files should have the same extension like the original
 ;;   files.  Strange behaviour with jka-compr otherwise?
 ;; * Copy files in dired from SMB to another method doesn't work.
 ;; * Try to remove the inclusion of dummy "" directory.  Seems to be at
-;;   several places, now.
+;;   several places, especially in `tramp-smb-handle-insert-directory'.
 ;; * Provide variables for debug.
 ;; * (RMS) Use unwind-protect to clean up the state so as to make the state
 ;;   regular again.
 
+;;; arch-tag: fcc9dbec-7503-4d73-b638-3c8aa59575f5
 ;;; tramp-smb.el ends here
